'******************************************************************************
' C Functions used by ed.spin
' Version 0.12
' Copyright (c) 2011, Dave Hein
' See end of file for terms of use.
'******************************************************************************
{{
  This object provides a subset of the CLIB object that is used by the ED
  text editor.  It can be configured to run under SpinSim using the conio and
  fileio objects, or it can run on actual hardware using the FullDuplexSerial
  and FSRW objects.
}}
CON
  ' SD Card Pin Assignments
  DO  = 16
  CLK = 21
  DI  = 20
  CS  = 19

OBJ
{
  ' SpinSim Objects
  ser : "conio"
  fio : "fileio"
}

  ' Hardware Objects
  ser : "FullDuplexSerial"
  fio : "fsrw"

pub start
{{ Start the serial and file I/O drivers }}
  ser.start(31, 30, 0, 115200)
  \fio.mount_explicit(DO, CLK, DI, CS)
  
pub strcpy(dst, src)
{{ Copy the source string to the destination string }}
  bytemove(dst, src, strsize(src) + 1)

pub strcat(dst, src)
{{ Concatenate the source string to the destination string }}
  bytemove(dst + strsize(dst), src, strsize(src) + 1)

pub memcpy(dst, src, num)
{{ Copy "num" bytes from "src" to "dst" }}
  bytemove(dst, src, num)

PUB isdigit(char)
{{ Return true if the character is an ASCII digit between "0" and "9" }}
  return char => "0" and char =< "9"

PUB strncmp(str1, str2, n)
{{ Compare the first "n" characters in "str1" with "str2" }}
  if (n =< 0)
    return 0
  repeat while (--n and byte[str1] and byte[str1] == byte[str2])
    str1++
    str2++
  return byte[str1] - byte[str2]

PUB malloc(size)
{{ Allocate a chunk of memory from the top of RAM }}
  result := $8000 - size

PUB sscanf1(str, fmt, ptr) | negate
{{ This routine implements a subset of the C sscanf routine }}
  ifnot strcomp(fmt, string("%d"))
    long[ptr] := 0
    return
  negate := 0
  if byte[str] == "-"
    negate := 1
    str++
  repeat while isdigit(byte[str])
    result := (result * 10) + byte[str++] - "0"
  if negate
    result := -result
  long[ptr] := result

pub printf0(fmt)
{{ Print a formatted string }}
  vprintf(fmt, 0)

pub printf1(fmt, parm1)
{{ Print a formatted string with one parameter }}
  vprintf(fmt, @parm1)

pub printf2(fmt, parm1, parm2)
{{ Print a formatted string with two parameters }}
  vprintf(fmt, @parm1)

pub printf3(fmt, parm1, parm2, parm3)
{{ Print a formatted string with three parameters }}
  vprintf(fmt, @parm1)

pub printf4(fmt, parm1, parm2, parm3, parm4)
{{ Print a formatted string with four parameters }}
  vprintf(fmt, @parm1)

pub vprintf(fmt, ptr) | val, digits
{{ This routine implements a subset of the C vprintf function }}
  repeat while byte[fmt]
    val := byte[fmt++]
    if val == "\"
      val := byte[fmt++]
      if val == "n"
        ser.tx(13)
      else
        ser.tx(val)
    elseif val == "%"
      digits := 0
      repeat while isdigit(byte[fmt]) or byte[fmt] == "."
        if byte[fmt] <> "."
          digits := byte[fmt] - "0"
        fmt++
      val := byte[fmt++]
      if val == "d"
        printpad(long[ptr], digits)
        ser.dec(long[ptr])
        ptr += 4
      elseif val == "x"
        ifnot digits
          digits := 8
        ser.hex(long[ptr], digits)
        ptr += 4
      elseif val == "s"
        ser.str(long[ptr])
        ptr += 4
      elseif val == "c"
        ser.tx(long[ptr])
        ptr += 4
      else
        ser.tx("%")
        ser.tx(val)
    else
      ser.tx(val)

PRI printpad(val, digits)
{{ This routine prints leading blanks depending on the values of "val" and "digits" }}
  if val < 0
    val := -val
    digits--
  repeat while digits > 1 and val > 9
    digits--
    val /= 10
  repeat while digits > 1
    ser.tx(" ")
    digits--

PUB gets(str) | char, str0
{{ Read a string from the serial input and edit with backspaces }}
  str0 := str
  repeat
    char := ser.rx
    if (char == 8)
      if (str > str0)
        ser.tx(8)
        ser.tx(" ")
        ser.tx(8)
        str--
      next
    if char == 4
      if (str == str0)
        str0 := 0
      quit
    ser.tx(char)
    if char == 10 OR char == 13
      quit
    else
      byte[str++] := char
  byte[str] := 0
  return str0

PUB fopen(fname, mode)
{{ Open the file "fname" using the first character in the "mode" string }}
  result := not \fio.popen(fname, byte[mode])

PUB fclose(stream)
{{ Close the file if "stream" is nonzero }}
  if stream
    result := not \fio.pclose

PUB fgets(str, num, stream)
{{ Read a string of up to "num" characters from a file }}
  ifnot stream
    return
  num-- ' Leave room for terminating null
  repeat while result < num
    if \fio.pread(str, 1) <> 1
      quit
    result++
    if byte[str++] == 10
      quit
  byte[str] := 0

PUB fprintf1(stream, fmt, parm1) | val
{{ This routine implements a subset of the C fprintf function }}
  ifnot stream
    return
  repeat while byte[fmt]
    val := byte[fmt++]
    if val == "%"
      val := byte[fmt++]
      if val == 0
        val := "%"
        \fio.pwrite(@val, 1)
        quit
      elseif val == "s"
        \fio.pwrite(parm1, strsize(parm1))
      else
        val |= "%" << 8
        \fio.pwrite(@val, 2)
    elseif val == "\"
      val := byte[fmt++]
      if val == 0
        val := "\"
        \fio.pwrite(@val, 1)
        quit
      elseif val == "n"
        val := $0a0d
        \fio.pwrite(@val, 2)
      else
        \fio.pwrite(@val, 1)
    else
      \fio.pwrite(@val, 1)

{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}